; ATtiny13 based Tesla coil protection (Undervoltage lockout and duty cycle limiter)
;
; Fuse settings:
; High fuse: 0xFB
; Low fuse: 0x7A
; 
; Clock source: internal RC oscillator at 9.6 MHz, SUT: 14 cycles + 64 ms
; Brown-out detection: Set at 2.7V
; SPI programming: enabled
;
; Copyright (c) 2016-2017, Frantisek Štefanec
;
; Firmware version 1.1
;

.include "tn13def.inc"


; ----- USER DEFINED VALUES -----

; Duty limiter time constant is set by hardware: resistor between PB0 and PB4, capacitor between PB4 and ground
; Duty cycle limiting is implemented on an external RC circuit, the percentages may not be accurate for lower BPS.
; Duty limiter values in %
	
	.equ MAX_DUTY=30
	.equ RELEASE_DUTY=20

; Lockout/release/MCU power voltage definitions in mV

	.equ UVLO_HIGH=9000
	.equ UVLO_LOW=8500

	.equ PWR_VOLTAGE=5000

; Overvoltage protection kick-in value in mV

	.equ OVLO_VALUE=14500

; Voltage divider resistor values (for supply voltage measurement)

	.equ VOLT_DIV_HIGH=10000
	.equ VOLT_DIV_LOW=2200

; Output and input pins

	.equ LED_PIN=PB3
	.equ OUT_PIN=PB0
	.equ INT_PIN=PB1


; ----- END OF USER DEFINED VALUES -----

; ------------------------------
; Register names

.def REG=R16
.def AD_STAT=R17
.def PINB_BUF=R18
.def ERR_V=R19
.def ERR_D=R20


; ------------------------------
; ADC release/lockout values

.equ VH_ADC_VALUE=(256*(UVLO_HIGH*VOLT_DIV_LOW/(VOLT_DIV_HIGH+VOLT_DIV_LOW))/PWR_VOLTAGE)
.equ VL_ADC_VALUE=(256*(UVLO_LOW*VOLT_DIV_LOW/(VOLT_DIV_HIGH+VOLT_DIV_LOW))/PWR_VOLTAGE)
.equ DCL_ADC_VALUE=(256*MAX_DUTY/100)
.equ DCL_REL_ADC_VALUE=(256*RELEASE_DUTY/100)
.equ OVLO_ADC_VALUE=(256*(OVLO_VALUE*VOLT_DIV_LOW/(VOLT_DIV_HIGH+VOLT_DIV_LOW))/PWR_VOLTAGE)

; ADLAR = result will be left-adjusted (top byte in ADCH), don't change
; ADC pins:

.equ ADMUX_DUTY=(1<<ADLAR)|(1<<MUX1)|(0<<MUX0) ; ADC2/PB4
.equ ADMUX_VOLT=(1<<ADLAR)|(0<<MUX1)|(1<<MUX0) ; ADC1/PB2

; ------------------------------
; Code segment

.CSEG

; ISR addresses - no ISRs here, just jump to main.

.ORG 0x00
	RJMP MAIN
.ORG 0x01
	NOP
.ORG 0x02
	NOP
.ORG 0x03
	NOP
.ORG 0x04
	NOP
.ORG 0x05
	NOP
.ORG 0x06
	NOP
.ORG 0x07
	NOP
.ORG 0x08
	NOP
.ORG 0x09
	NOP
.ORG 0x0A

MAIN:
	; clear interrupts
	CLI

	; set stack pointer
	LDI 	REG, LOW(RAMEND)
	OUT 	SPL, REG

	; set outputs and inputs
	LDI 	REG, (1<<OUT_PIN)|(1<<LED_PIN)
	OUT 	DDRB, REG
	LDI	REG, 0
	OUT	PORTB, REG
	
	; check if watchdog reset happened - if yes, freeze and blink SOS
	IN	REG, MCUSR
	ANDI	REG, (1<<WDRF)
	BREQ 	NO_WDR
	LDI	REG, 0
	OUT	MCUSR, REG
	RJMP 	FREEZE_BLINK_SOS

	; flash LED quickly & wait for 600 ms before doing anything else (time for voltage to stabilize)
	SBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	SBI 	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	CBI 	PORTB, LED_PIN
	RCALL 	DLY_100ms
	SBI 	PORTB, LED_PIN


NO_WDR:
	
	; disable watchdog
	RCALL 	DISABLE_WD

	; set up AD converter
	LDI	REG, (1<<ADEN)|(1<<ADPS0)|(1<<ADPS0)|(1<<ADPS2) ; enable ADC, set :128 prescaler - ADC is running at ~750 kHz, sample rate ~6 kSPS
	OUT 	ADCSRA, REG
	LDI	REG, ADMUX_VOLT
	OUT 	ADMUX, REG
	LDI 	AD_STAT, 0

	; wait 200 ms before doing anything else (time for voltages to stabilize)
	RCALL	DLY_100ms
	RCALL	DLY_100ms

	; check supply voltage (UVLO)
	RCALL	AD_CONV
	IN	REG, ADCH
	CPI	REG, VH_ADC_VALUE
	BRSH	START_VOLT_OKL
	RCALL 	VOLT_LIMIT
START_VOLT_OKL:
	; check supply voltage (OVLO)
	CPI	REG, OVLO_ADC_VALUE
	BRLO	START_VOLT_OKH
	RCALL	FREEZE_BLINK_SOS
START_VOLT_OKH:

	; start another AD conversion
	SBI	ADCSRA,	ADSC

	; enable LED
	SBI	PORTB, LED_PIN

	; call the Enable Watchdog subroutine
	; The microcontroller is running near strong electromagnetic fields, in case it crashes and keeps its output high,
	; the watchdog should restart it to prevent potential damage to the power electronics.
	; Upon restart, the indicator LED will blink with an "SOS" patern.
	RCALL 	ENABLE_WD

LOOP:	; main program loop
	; "pat" the watchdog
	WDR

	; load Port B value into register
	IN	PINB_BUF, PINB

	; transfer interrupter pin value from register to output pin
	SBRS 	PINB_BUF, INT_PIN
	CBI	PORTB, OUT_PIN
	SBRC	PINB_BUF, INT_PIN
	SBI	PORTB, OUT_PIN

 	; skip if conversion in progress - just repeat loop
	SBIC	ADCSRA, ADSC
	RJMP	LOOP
	
	; if ADC conversion is finished:
	; were we measuring duty cycle?

	CPI	AD_STAT, 1
	BRNE 	DCL_FIN
	IN	REG, ADCH
	CPI	REG, DCL_ADC_VALUE
	BRLO	DT_OK ; if duty cycle lower than maximum, OK
	INC	ERR_D
	CPI	ERR_D, 4
	BRLO	DT_ER1
	RCALL	DUTY_LIMIT ; else call DCL routine
	RJMP	ADC_FIN
DT_OK:	CLR	ERR_D
DT_ER1:	LDI	AD_STAT, 0 
	LDI	REG, ADMUX_VOLT
	OUT	ADMUX, REG
	SBI	ADCSRA, ADSC
	RJMP 	ADC_FIN
DCL_FIN:

	; were we measuring supply voltage?
			
	CPI	AD_STAT, 0
	BRNE	ADC_FIN	
	IN	REG, ADCH
	CPI	REG, VL_ADC_VALUE
	BRSH	V_OKL ; if voltage higher or same as minimum, OK
	INC	ERR_V
	CPI	ERR_V, 4
	BRLO	V_ER1
	RCALL	VOLT_LIMIT ; else call UVLO routine
	RJMP 	ADC_FIN
V_OKL:	CLR	ERR_V	
V_ER1:	CPI	REG, OVLO_ADC_VALUE	
	BRLO	V_OKH
	RCALL	FREEZE_BLINK_SOS
V_OKH:	LDI	AD_STAT, 1
	LDI	REG, ADMUX_DUTY
	OUT	ADMUX, REG
	SBI	ADCSRA, ADSC 

ADC_FIN:
	RJMP LOOP


DLY_100ms:
	; 100 ms delay loop at 9.6MHz
	; generated using http://www.bretmulvey.com/avrdelay.html
	PUSH	R18
	PUSH	R19
	PUSH	R20
	LDI	R18, 5
	LDI	R19, 223
	LDI	R20, 182
L1:	DEC	R20
	BRNE	L1
	DEC 	R19
	BRNE	L1
	DEC	R18
	BRNE 	L1
	POP 	R20
	POP	R19
	POP	R18
	RET
	
AD_CONV:
	; start Analog to digital conversion
	SBI	ADCSRA, ADSC
AD_WAIT:
	; wait while running, then return from subroutine
	SBIC 	ADCSRA, ADSC
	RJMP 	AD_WAIT
	RET

	; Undervoltage LockOut routine
VOLT_LIMIT:
	CBI	PORTB, OUT_PIN
	RCALL	DISABLE_WD
VOLT_LIMIT_REPEAT: 
	; 1 s wait loop with blinking led with 1/5 duty and 500 ms period
	SBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	SBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	RCALL	DLY_100ms
	; measure supply voltage again
	LDI	REG, ADMUX_VOLT
	OUT	ADMUX, REG
	RCALL	AD_CONV
	IN	REG, ADCH
	CPI	REG, VH_ADC_VALUE
	BRLO	VOLT_LIMIT_REPEAT
	; if still lower, repeat the wait cycle and supply voltage measurement
	; if OK, set measurement to duty instead and return
	LDI	AD_STAT, 1
	LDI	REG, ADMUX_DUTY
	OUT	ADMUX, REG
	SBI	ADCSRA, ADSC 
	RCALL	ENABLE_WD
	SBI	PORTB, LED_PIN
	RET
	
	; Duty Cycle Limiter routine
DUTY_LIMIT:
	CBI	PORTB, OUT_PIN
	RCALL	DISABLE_WD
	; 400 ms wait loop with blinking led with 1/2 duty and 200 ms period
DUTY_LIMIT_REPEAT:
	SBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL	DLY_100ms
	; measure duty limiter voltage again
	LDI	REG, ADMUX_DUTY
	OUT	ADMUX, REG
	RCALL	AD_CONV
	IN	REG, ADCH
	; if still too high, repeat the wait cycle and duty measurement
	CPI	REG, DCL_REL_ADC_VALUE
	BRSH	DUTY_LIMIT_REPEAT
	; if interrupter input pin in ON, repeat the wait cycle
	IN 	PINB_BUF, PINB
	SBRC	PINB_BUF, INT_PIN
	RJMP	DUTY_LIMIT_REPEAT	
	; if OK, set measurement to voltage instead and return
	LDI	AD_STAT, 0
	LDI	REG, ADMUX_VOLT
	OUT	ADMUX, REG
	SBI	ADCSRA, ADSC 
	RCALL	ENABLE_WD
	SBI	PORTB, LED_PIN
	RET
	
	; Watchdog disable
DISABLE_WD:
	WDR
	; clear WDRF in MCUSR
	IN	REG, MCUSR
	ANDI	REG, (0xFF - (1<<WDRF))
	OUT	MCUSR, REG
	; watchdog change enable
	IN	REG, WDTCR
	ORI	REG, (1<<WDCE)|(1<<WDE)
	OUT	WDTCR, REG
	; turn off WDT
	LDI	REG, (0<<WDE)
	OUT	WDTCR, REG
	RET

	; Watchdog enable
ENABLE_WD:
	WDR
	; watchdog change enable
	IN	REG, WDTCR
	ORI	REG, (1<<WDCE)|(1<<WDE)
	OUT	WDTCR, REG
	; set prescaler
	LDI	REG, (1<<WDE)|(1<<WDP0) ; watchdog timeout = 16 ms
	OUT	WDTCR, REG
	RET

; subroutine to freeze and blink with a SOS pattern on microcontroller crash

FREEZE_BLINK_SOS:
	; disable output pin and watchdog
	CBI	PORTB, OUT_PIN
	RCALL	DISABLE_WD

	; 3 short blinks
	LDI	REG, 3
LP1:	SBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	DEC	REG
	BRNE	LP1

	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms

	; 3 long blinks
	LDI	REG, 3
LP2:	SBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	DEC	REG
	BRNE	LP2
	
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms

	; 3 short blinks
	LDI	REG, 3
LP3:	SBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	CBI	PORTB, LED_PIN
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	DEC	REG
	BRNE	LP3

	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms
	RCALL 	DLY_100ms

	RJMP 	FREEZE_BLINK_SOS
